@@ -164,3 +164,42 @@ test!(
     false
     "#
 );
+
+test!(
+    ascii_downcase,
+    r#"
+    ascii_downcase
+    "#,
+    r#"
+    "abCdEF123"
+    "#,
+    r#"
+    "abcdef123"
+    "#
+);
+
+test!(
+    ascii_upcase,
+    r#"
+    ascii_upcase
+    "#,
+    r#"
+    "abCdEF123"
+    "#,
+    r#"
+    "ABCDEF123"
+    "#
+);
+
+test!(
+    tailrec_opt,
+    r#"
+    def walk2(f): if type | . == "array" or . == "object" then map_values(walk2(f)) | f else f end; walk2(.)
+    "#,
+    r#"
+    [1,[2,3]]
+    "#,
+    r#"
+    [1,[2,3]]
+    "#
+);
